#include <getopt.h>
#include <asm/types.h>
#include <assert.h>
#include <fcntl.h>
#include <dirent.h>
#include <malloc.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>

#include <linux/stddef.h>
#include <linux/videodev2.h>
#include <errno.h>

#include <rga/rga.h>
#include <rga/im2d.h>
#include <rockchip/rockchip_drm.h>

#define CMA_ALLOC_ENABLE	1

#define DEFAULT_WH		"3840x2160"
#define DEFAULT_FORMAT		"nv12"
#define DEFAULT_CROP		"0,0,0,0"

unsigned long long nanoTime()
{
    struct timespec t;

    t.tv_sec = t.tv_nsec = 0;
    clock_gettime(CLOCK_MONOTONIC, &t);

    return (unsigned long long)(t.tv_sec) * 1000000000 + t.tv_nsec;
}

static void help(void)
{
    printf(
        "Usage: rgatest [Options]\n\n"
        "Options:\n"
	"-i, --input                  Input file, default is /dev/zero\n"
	"-o, --output                 Output file, default is empty\n"
        "-s, --src                    Src image, such as 1920x1080@0,0,1920,1080@nv12, default is 3840x2160@0,0,0,0@nv12\n"
        "-d, --dst                    Dst image, such as 1920x1080@0,0,1280,720@rgb24, default is 3840x2160@0,0,0,0@nv12\n"
        "-r, --rotate                 Image rotation degree, the value should be 90, 180 or 270\n"
        "-V, --vflip                  Vertical Mirror\n" 
        "-H, --hflip                  Horizontal Mirror\n"
        "\n"
        "e.g.\n"
        "Image copy:                    rgatest -s 3840x2160 -d 3840x2160\n"
        "Image format:                  rgatest -s 3840x2160@nv12 -d 3840x2160@rgb24\n"
        "Image scale:                   rgatest -s 3840x2160 -d 1920x1080\n"
        "Image rotate:                  rgatest -s 3840x2160 -d 3840x2160 -r 90\n"
        "Image crop:                    rgatest -s 3840x2160 -d 3840x2160:0,0,1920,1080\n"
        "Image format and crop:         rgatest -s 3840x2160@nv12 -d 3840x2160@0,0,1920,1080@rgb24\n"
        "Image format, crop and rotate: rgatest -s 3840x2160@nv12 -d 3840x2160@0,0,1920,1080@rgb24 -r 90\n"
        "Image format and write:        rgatest -s 3840x2160@nv12 -d 3840x2160@rgb24 -i test.nv12 -o test.rgb24\n"
        "\n");
}

static const char *short_options = "i:o:s:d:r:VHc:h";

static struct option long_options[] = {
    {"input", required_argument, NULL, 'i'},
    {"output", required_argument, NULL, 'o'},
    {"src", required_argument, NULL, 's'},
    {"dst", required_argument, NULL, 'd'},
    {"rotate", required_argument, NULL, 'r'},
    {"vflip", no_argument, NULL, 'V'},
    {"hflip", no_argument, NULL, 'H'},
    {"help", no_argument, NULL, 'h'},
    {NULL, 0, NULL, 0}
};

static unsigned int get_usage(int rotate, int vflip, int hflip)
{
    unsigned int usage = IM_SYNC;
    
    switch (rotate) {
    case 90: usage |= IM_HAL_TRANSFORM_ROT_90; break;
    case 180: usage |= IM_HAL_TRANSFORM_ROT_180; break;
    case 270: usage |= IM_HAL_TRANSFORM_ROT_270; break;
    default: break;
    }

    if (vflip)
        usage |= IM_HAL_TRANSFORM_FLIP_V;

    if (hflip)
        usage |= IM_HAL_TRANSFORM_FLIP_H;

    return usage;
}

static DrmBuffer *parse_image_parameters(char *str, int drm_fd, im_rect *rect, rga_buffer_t *rbuf, int rotate, int is_src)
{
    int i = 0;
    char *p;
    char str_format[32], str_wh[32], str_crop[32]; 
    int wh[2];
    int crop[4];
    RgaSURF_FORMAT rgaformat;
    unsigned int v4l2format;
    DrmBuffer *cma_buf = NULL;

    memset(wh, 0, 2 * sizeof(int));
    memset(crop, 0, 4 * sizeof(int));

    strcpy(str_wh, DEFAULT_WH);
    strcpy(str_crop, DEFAULT_CROP);
    strcpy(str_format, DEFAULT_FORMAT);

    i = 0;
    p = strtok(str, "@");
    while(p != NULL) {
        if (strstr(p, "x") != NULL)
            strcpy(str_wh, p);
        else if (strstr(p, ",") != NULL)
            strcpy(str_crop, p);
        else
            strcpy(str_format, p);
        p = strtok(NULL, "@");

        if(i >= 3)
             break;
    }

    i = 0;
    p = strtok((char *)str_wh, "x");
    while(p != NULL) {
        wh[i++] = atoi(p);
        p = strtok(NULL, "x");

        if(i >= 2)
             break;
    }

    i = 0;
    p = strtok((char *)str_crop, ",");
    while(p != NULL) {
        crop[i++] = atoi(p);
        p = strtok(NULL, ",");

        if(i >= 4)
             break;
    }

    rect->x = crop[0];
    rect->y = crop[1];
    rect->width = crop[2];
    rect->height = crop[3];

    if (strcmp(str_format, "nv12") == 0 || strcmp(str_format, "NV12") == 0) {
        rgaformat = RK_FORMAT_YCbCr_420_SP;
        v4l2format = V4L2_PIX_FMT_NV12;
    } else if (strcmp(str_format, "rgb24") == 0 || strcmp(str_format, "RGB24") == 0) {
        rgaformat = RK_FORMAT_RGB_888; 
        v4l2format = V4L2_PIX_FMT_RGB24;
    } else if (strcmp(str_format, "bgr24") == 0 || strcmp(str_format, "BGR24") == 0) {
        rgaformat = RK_FORMAT_RGB_888;
        v4l2format = V4L2_PIX_FMT_BGR24;
    } else {
        rgaformat = RK_FORMAT_YCbCr_420_SP;
        v4l2format = V4L2_PIX_FMT_NV12;
    }

#ifdef CMA_ALLOC_ENABLE
    cma_buf = RockchipCmaAlloc(drm_fd, v4l2format, wh[0], wh[1]);
#else
    cma_buf = RockchipDrmAlloc(drm_fd, v4l2format, wh[0], wh[1]);
#endif
    if (cma_buf == NULL) {
        printf("No memory for rga\n");
        return NULL;
    }


    if (!is_src && (rotate == 90 || rotate == 270)) {
        *rbuf = wrapbuffer_fd_t(cma_buf->fd, DRM_ALIGN(wh[1], 16), wh[0], DRM_ALIGN(wh[1], 16), wh[0], rgaformat);
        if (rect->width == 0 || rect->height == 0) {
            rect->width = wh[1];
            rect->height = wh[0];
        }
    } else {
        *rbuf = wrapbuffer_fd_t(cma_buf->fd, wh[0], wh[1], DRM_ALIGN(wh[0], 16), wh[1], rgaformat);
    }
    
    printf("width %d, height %d, crop %d,%d,%d,%d\n", wh[0], wh[1], crop[0], crop[1], crop[2], crop[3]);
    return cma_buf;
}

int rga_convert(char *in, char *out, char *src_fmt, char *dst_fmt, int rotate, int vfilp, int hflip)
{

    int ret = 0;
    FILE *fp_in = NULL;
    FILE *fp_out = NULL;
    int drm_fd = -1;
    DrmBuffer *src_cmabuf = NULL, *dst_cmabuf = NULL;
    unsigned long long start, stop;

    uint32_t usage = 0;
    im_rect src_rect, dst_rect, pat_rect;
    rga_buffer_t src_rgabuf, dst_rgabuf, pat_rgabuf; 

    memset(&src_rect, 0, sizeof(im_rect));
    memset(&dst_rect, 0, sizeof(im_rect));
    memset(&pat_rect, 0, sizeof(im_rect));

    memset(&src_rgabuf, 0, sizeof(rga_buffer_t));
    memset(&dst_rgabuf, 0, sizeof(rga_buffer_t));
    memset(&pat_rgabuf, 0, sizeof(rga_buffer_t));

    drm_fd = RockchipDrmOpen();
    if (drm_fd < 0) {
        ret = drm_fd;
        goto err;
    }

    src_cmabuf = parse_image_parameters(src_fmt, drm_fd, &src_rect, &src_rgabuf, rotate, 1);
    if (!src_cmabuf) {
        ret = -EINVAL;
        goto err;
    }
    
    dst_cmabuf = parse_image_parameters(dst_fmt, drm_fd, &dst_rect, &dst_rgabuf, rotate, 0);
    if (!src_cmabuf) {
        ret = -EINVAL;
        goto err;
    }
    
    usage = get_usage(rotate, vfilp, hflip);

    if (in != NULL)
        fp_in = fopen(in, "r");
    if (out != NULL)
        fp_out = fopen(out, "w+");

    if (fp_in) {
        while(fread(src_cmabuf->virt, 1, src_cmabuf->size, fp_in) == src_cmabuf->size && !feof(fp_in)) {
            IM_STATUS status;    

            start = nanoTime();
            status = improcess(src_rgabuf, dst_rgabuf, pat_rgabuf, src_rect, dst_rect, pat_rect, usage);
            if (status <= 0) {
                printf("RGA process failed, status %d\n", status);
                ret = -EIO;
                break;
            }
            stop = nanoTime();
            if (fp_out)
                fwrite(dst_cmabuf->virt, 1, dst_cmabuf->size, fp_out);
            printf("RGA process time: %lldms\n", (stop - start) / 1000000);
        }
    }

err:
    if (fp_in)
        fclose(fp_in);

    if (fp_out)
        fclose(fp_out);

    if (src_cmabuf)
#ifdef CMA_ALLOC_ENABLE
        RockchipCmaFree(drm_fd, src_cmabuf);
#else
        RockchipDrmFree(drm_fd, src_cmabuf);
#endif

    if (dst_cmabuf)
#ifdef CMA_ALLOC_ENABLE
        RockchipCmaFree(drm_fd, dst_cmabuf);
#else
        RockchipDrmFree(drm_fd, dst_cmabuf);
#endif

    if (drm_fd >= 0)
        RockchipDrmClose(drm_fd);
    
    return ret;
}

int main(int argc, char **argv)
{
    int c, r;
    char index = -1;
    char *src = NULL, *dst = NULL, *in = NULL, *out = NULL;
    int vflip = 0;
    int hflip = 0;
    int rotate = 0;

    if (argc == 1) {
        help();
        return -EINVAL;
    }

    while((c = getopt_long(argc, argv, short_options, long_options, NULL)) != -1) {
        switch (c) {
        case 'i':
            in = optarg;
            break;
        case 'o':
            out = optarg;
            break;
        case 's':
            src = optarg;
            break;
        case 'd':
            dst = optarg;
            break;
        case 'r':
            rotate = atoi(optarg);
            break;
        case 'V':
            vflip = 1;
            break;
        case 'H':
            hflip = 1;
            break;
        case 'h':
            help();
            return 0;
        default:
            help();
            return -EINVAL;
        }
    }

    if (in == NULL)
        in = "/dev/zero";

    printf("in %s, out %s, src %s, dst %s, rotate %d, hflip %d, vflip %d\n", in, out, src, dst, rotate, hflip, vflip);
    return rga_convert(in, out, src, dst, rotate, vflip, hflip);
}
